# Copyright © 2022 Gurobi Optimization, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Module for inserting simple Scikit-Learn regression models into a gurobipy model.

All linear models should work:
   - :external+sklearn:py:class:`sklearn.linear_model.LinearRegression`
   - :external+sklearn:py:class:`sklearn.linear_model.Ridge`
   - :external+sklearn:py:class:`sklearn.linear_model.Lasso`
"""


from ..modeling import AbstractPredictorConstr
from .skgetter import SKgetter


class BaseSKlearnRegressionConstr(SKgetter, AbstractPredictorConstr):
    """Predict a Gurobi variable using a Linear Regression that
    takes another Gurobi matrix variable as input.
    """

    def __init__(
        self,
        gp_model,
        predictor,
        input_vars,
        output_vars=None,
        output_type="",
        **kwargs,
    ):
        self._output_shape = 1
        SKgetter.__init__(self, predictor, input_vars, output_type, **kwargs)
        AbstractPredictorConstr.__init__(
            self,
            gp_model,
            input_vars,
            output_vars,
            **kwargs,
        )

    def add_regression_constr(self, output=None):
        """Add the prediction constraints to Gurobi."""
        if output is None:
            output = self.output
        coefs = self.predictor.coef_.reshape(-1, 1)
        intercept = self.predictor.intercept_
        self.gp_model.addConstr(output == self.input @ coefs + intercept, name="linreg")



# Copyright © 2022 Gurobi Optimization, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from abc import ABC, abstractmethod

import gurobipy as gp

from ..exceptions import ParameterError
from ._var_utils import _get_sol_values, validate_input_vars, validate_output_vars
from .submodel import SubModel


class AbstractPredictorConstr(ABC, SubModel):
    """Base class to store sub-model added by :py:func:`gurobi_ml.add_predictor_constr`.

    This class is the base class to store everything that is added to
    a Gurobi model when a trained predictor is inserted into it. Depending on
    the type of the predictor, a class derived from it will be returned
    by :py:func:`gurobi_ml.add_predictor_constr`.

    Warning
    -------

    Users should usually never construct objects of this class or one of its derived
    classes. They are returned by the :py:func:`gurobi_ml.add_predictor_constr` and
    other functions.
    """

    def __init__(self, gp_model, input_vars, output_vars=None, **kwargs):
        self._input = input_vars
        self._output = output_vars
        SubModel.__init__(self, gp_model, **kwargs)

    def _validate(self):
        """Validate input and output variables (check shapes, reshape if needed)."""
        input_vars = self.input
        output_vars = self.output

        if hasattr(self, "_input_shape") and input_vars.shape[1] != self._input_shape:
            raise ParameterError(
                "Input variables dimension doesn't conform with modeling object "
                + f"{type(self)}, input variable dimension: "
                + f"{self._input_shape} != {input_vars.shape[1]}"
            )

        if output_vars.ndim == 1:
            if input_vars.shape[0] == 1:
                output_vars = output_vars.reshape((1, -1))
            else:
                output_vars = output_vars.reshape((-1, 1))

        if (
            hasattr(self, "_output_shape")
            and output_vars.shape[1] != self._output_shape
        ):
            raise ParameterError(
                "Output variables dimension doesn't conform with modeling object "
                + f"{type(self)}, output variable dimension: "
                + f"{output_vars.shape[1]}"
            )

        if output_vars.shape[0] != input_vars.shape[0]:
            raise ParameterError(
                "Non-conforming dimension between "
                + "input variable and output variable: "
                + f"{output_vars.shape[0]} != {input_vars.shape[0]}"
            )

        self._input = input_vars
        self._output = output_vars

    def _build_submodel(self, gp_model, *args, **kwargs):
        """Predict output from input using predictor or transformer."""
        self._input, columns, index = validate_input_vars(self._gp_model, self._input)
        self._input_index = index
        self._input_columns = columns

        if self._output is None:
            self._output = self._create_output_vars(self._input)
        if self._output is not None:
            self._output = validate_output_vars(self._output)
            self._validate()
        self._mip_model(**kwargs)
        assert self._output is not None
        return self

    def _print_container_steps(self, iterations_name, iterable, file):
        header = f"{iterations_name:13} {'Output Shape':>14} {'Variables':>12} {'Constraints':^38}"
        print("-" * len(header), file=file)
        print(header, file=file)
        print(f"{' '*41} {'Linear':>12} {'Quadratic':>12} {'General':>12}", file=file)
        print("=" * len(header), file=file)
        for step in iterable:
            step.print_stats(abbrev=True, file=file)
            print(file=file)
        print("-" * len(header), file=file)

    def print_stats(self, abbrev=False, file=None):
        """Print statistics on model additions stored by this class.

        This function prints detailed statistics on the variables
        and constraints that where added to the model.

        Usually derived classes reimplement this function to provide more
        details about the structure of the additions (type of ML model,
        layers if it's a neural network,...)

        Arguments
        ---------

        file: None, optional
            Text stream to which output should be redirected. By default sys.stdout.
        """

        if abbrev:
            print(
                f"{self._name:13} {self.output.shape.__str__():>14} {len(self.vars):>12} "
                + f"{len(self.constrs):>12} {len(self.qconstrs):>12} {len(self.genconstrs):>12}",
                file=file,
            )
        else:
            super().print_stats(file)
            print(f"Input has shape {self.input.shape}", file=file)
            print(f"Output has shape {self.output.shape}", file=file)

    def _create_output_vars(self, input_vars, name="out"):
        """May be defined in derived class to create the output variables of predictor."""
        try:
            n_outputs = self._output_shape
        except AttributeError:
            return
        output = self._gp_model.addMVar(
            (input_vars.shape[0], n_outputs),
            lb=-gp.GRB.INFINITY,
            name=self._name_var(name),
        )
        self._gp_model.update()
        return output

    @property
    def _has_solution(self):
        """Returns true if we have a solution."""
        try:
            self.input_values
            return True
        except gp.GurobiError:
            pass
        return False

    @abstractmethod
    def get_error(self):
        """Returns error in Gurobi's solution with respect to prediction from input.

        Returns
        -------
        error : ndarray of same shape as :py:attr:`gurobi_ml.modeling.base_predictor_constr.AbstractPredictorConstr.output`
            Assuming that we have a solution for the input and output variables
            `x, y`. Returns the absolute value of the differences between `predictor.predict(x)` and
            `y`. Where predictor is the regression model represented by this object.

        Raises
        ------
        NoSolution
            If the Gurobi model has no solution (either was not optimized or is infeasible).
        """
        ...

    @abstractmethod
    def _mip_model(self, **kwargs):
        """Makes MIP model for the predictor the sub-class implements."""
        ...

    def _indexed_name(self, index, name):
        index = f"{index}".replace(" ", "")
        return self._name_var(f"{name}[{index}]")

    def _name_var(self, name):
        if self._name != "" and not self._no_recording:
            return name
        return None

    @property
    def output(self):
        """Output variables of embedded predictor.

        Returns
        -------
        output : :gurobipy:`MVar`.
        """
        return self._output

    @property
    def input_values(self):
        """Returns the values for the input variables if a solution is known.

        Returns
        -------
        output_value : ndarray or pandas dataframe with values

        Raises
        ------
        NoSolution
            If the Gurobi model has no solution (either was not optimized or is infeasible).
        """

        return _get_sol_values(self.input, self._input_columns, self._input_index)

    @property
    def output_values(self):
        """Returns the values for the output variables if a solution is known.

        Returns
        -------
        output_value : ndarray or pandas dataframe with values

        Raises
        ------
        NoSolution
            If the Gurobi model has no solution (either was not optimized or is infeasible).
        """
        return _get_sol_values(self.output)

    @property
    def input(self):
        """Returns the input variables of embedded predictor.

        Returns
        -------
        output : :gurobipy:`MVar`.
        """
        return self._input

    def __str__(self):
        return self._name



# Copyright © 2022 Gurobi Optimization, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Module for formulating a :external+sklearn:py:class:`sklearn.neural_network.MLPRegressor` in a
:gurobipy:`model`.
"""
from ..exceptions import NoModel
from ..modeling.neuralnet import BaseNNConstr
from .skgetter import SKgetter


def add_mlp_regressor_constr(
    gp_model, mlp_regressor, input_vars, output_vars=None, **kwargs
):
    """Formulate mlp_regressor in gp_model.

    The formulation predicts the values of output_vars using input_vars according to
    mlp_regressor. See our :ref:`Users Guide <Neural Networks>` for details on the mip
    formulation used.

    Parameters
    ----------
    gp_model : :gurobipy:`model`
        The gurobipy model where the predictor should be inserted.
    mlpregressor : :external+sklearn:py:class:`sklearn.neural_network.MLPRegressor`
        The multi-layer perceptron regressor to insert as predictor.
    input_vars : :gurobipy:`mvar` or :gurobipy:`var` array like
        Decision variables used as input for regression in model.
    output_vars : :gurobipy:`mvar` or :gurobipy:`var` array like, optional
        Decision variables used as output for regression in model.

    Returns
    -------
    MLPRegressorConstr
        Object containing information about what was added to gp_model to formulate
        mlp_regressor.

    Raises
    ------
    NoModel
        If the translation to Gurobi of the activation function for the network is not
        implemented.

    Note
    ----
    |VariablesDimensionsWarn|
    """
    return MLPRegressorConstr(
        gp_model, mlp_regressor, input_vars, output_vars, **kwargs
    )


class MLPRegressorConstr(SKgetter, BaseNNConstr):
    """Class to model trained
    :external+sklearn:py:class:`sklearn.neural_network.MLPRegressor` with gurobipy.

    |ClassShort|
    """

    def __init__(
        self,
        gp_model,
        predictor,
        input_vars,
        output_vars=None,
        clean_predictor=False,
        **kwargs,
    ):
        SKgetter.__init__(self, predictor, input_vars, **kwargs)
        BaseNNConstr.__init__(
            self,
            gp_model,
            predictor,
            input_vars,
            output_vars,
            clean_predictor=clean_predictor,
            **kwargs,
        )
        assert predictor.out_activation_ in ("identity", "relu")

    def _mip_model(self, **kwargs):
        """Add the prediction constraints to Gurobi."""
        neural_net = self.predictor
        if neural_net.activation not in self.act_dict:
            print(self.act_dict)
            raise NoModel(
                neural_net,
                f"No implementation for activation function {neural_net.activation}",
            )
        activation = self.act_dict[neural_net.activation]

        input_vars = self._input
        output = None

        for i in range(neural_net.n_layers_ - 1):
            layer_coefs = neural_net.coefs_[i]
            layer_intercept = neural_net.intercepts_[i]

            # For last layer change activation
            if i == neural_net.n_layers_ - 2:
                activation = self.act_dict[neural_net.out_activation_]
                output = self._output

            layer = self.add_dense_layer(
                input_vars,
                layer_coefs,
                layer_intercept,
                activation,
                output,
                **kwargs,
            )
            input_vars = layer._output  # pylint: disable=W0212
            self._gp_model.update()
        assert (
            self._output is not None
        )  # Should never happen since sklearn object defines n_ouputs_



# Copyright © 2022 Gurobi Optimization, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Module for inserting ordinary Scikit-Learn regression models into a
:gurobipy:`model`.

The following linear models are tested and should work:
   - :external+sklearn:py:class:`sklearn.linear_model.LinearRegression`
   - :external+sklearn:py:class:`sklearn.linear_model.Ridge`
   - :external+sklearn:py:class:`sklearn.linear_model.Lasso`
"""

from .base_regressions import BaseSKlearnRegressionConstr


def add_linear_regression_constr(
    gp_model, linear_regression, input_vars, output_vars=None, **kwargs
):
    """Formulate linear_regression in gp_model.

    The formulation predicts the values of output_vars using input_vars according to
    linear_regression. See our :ref:`Users Guide <Linear Regression>` for details on the
    mip formulation used.

    Parameters
    ----------
    gp_model : :gurobipy:`model`
        The gurobipy model where the predictor should be inserted.
    linear_regression : :external+sklearn:py:class:`sklearn.linear_model.LinearRegression`
     The linear regression to insert. It can be of any of the following types:
         * :external+sklearn:py:class:`sklearn.linear_model.LinearRegression`
         * :external+sklearn:py:class:`sklearn.linear_model.Ridge`
         * :external+sklearn:py:class:`sklearn.linear_model.Lasso`
     input_vars: :gurobipy:`mvar` or :gurobipy:`var` array like
         Decision variables used as input for random forest in model.
     output_vars: :gurobipy:`mvar` or :gurobipy:`var` array like, optional
         Decision variables used as output for random forest in model.

    Returns
    -------
    LinearRegressionConstr
        Object containing information about what was added to gp_model to formulate
        linear_regression.

    Note
    ----
    |VariablesDimensionsWarn|
    """
    return LinearRegressionConstr(
        gp_model, linear_regression, input_vars, output_vars, **kwargs
    )


class LinearRegressionConstr(BaseSKlearnRegressionConstr):
    """Class to model trained
    :external+sklearn:py:class:`sklearn.linear_model.LinearRegression` with gurobipy
    |ClassShort|.
    """

    def __init__(self, gp_model, predictor, input_vars, output_vars=None, **kwargs):
        self._default_name = "lin_reg"
        BaseSKlearnRegressionConstr.__init__(
            self,
            gp_model,
            predictor,
            input_vars,
            output_vars,
            **kwargs,
        )

    def _mip_model(self, **kwargs):
        """Add the prediction constraints to Gurobi."""
        self.add_regression_constr()



# Copyright © 2022 Gurobi Optimization, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Module for formulating :external+torch:py:class:`torch.nn.Sequential` model in a
:gurobipy:`model`.
"""

import numpy as np
import torch
from torch import nn

from ..exceptions import NoModel, NoSolution
from ..modeling.neuralnet import BaseNNConstr


def add_sequential_constr(
    gp_model, sequential_model, input_vars, output_vars=None, **kwargs
):
    """Formulate sequential_model into gp_model.

    The formulation predicts the values of output_vars using input_vars according to sequential_model.
    See our :ref:`Users Guide <Neural Networks>` for details on the mip formulation used.

    Parameters
    ----------
    gp_model : :gurobipy:`model`
        The gurobipy model where the sequential model should be inserted.
    sequential_model : :external+torch:py:class:`torch.nn.Sequential`
        The sequential model to insert as predictor.
    input_vars : :gurobipy:`mvar` or :gurobipy:`var` array like
        Decision variables used as input for logistic regression in model.
    output_vars : :gurobipy:`mvar` or :gurobipy:`var` array like, optional
        Decision variables used as output for logistic regression in model.

    Returns
    -------
    SequentialConstr
        Object containing information about what was added to model to insert the
        predictor in it

    Raises
    ------
    NoModel
        If the translation for some of the Pytorch model structure
        (layer or activation) is not implemented.

    Warning
    -------
    Only :external+torch:py:class:`torch.nn.Linear` layers and
    :external+torch:py:class:`torch.nn.ReLU` layers are supported.

    Note
    ----
    |VariablesDimensionsWarn|
    """
    return SequentialConstr(
        gp_model, sequential_model, input_vars, output_vars, **kwargs
    )


class SequentialConstr(BaseNNConstr):
    """Transform a pytorch Sequential Neural Network to Gurobi constraint with
    input and output as matrices of variables.
    |ClassShort|.
    """

    def __init__(self, gp_model, predictor, input_vars, output_vars=None, **kwargs):
        for step in predictor:
            if isinstance(step, nn.ReLU):
                pass
            elif isinstance(step, nn.Linear):
                pass
            else:
                raise NoModel(predictor, f"Unsupported layer {type(step).__name__}")
        super().__init__(gp_model, predictor, input_vars, output_vars)

    def _mip_model(self, **kwargs):
        network = self.predictor
        _input = self._input
        output = None
        num_layers = len(network)

        for i, step in enumerate(network):
            if i == num_layers - 1:
                output = self._output
            if isinstance(step, nn.ReLU):
                layer = self.add_activation_layer(
                    _input, self.act_dict["relu"], output, name=f"relu_{i}", **kwargs
                )
                _input = layer.output
            elif isinstance(step, nn.Linear):
                for name, param in step.named_parameters():
                    if name == "weight":
                        layer_weight = param.detach().numpy().T
                    elif name == "bias":
                        layer_bias = param.detach().numpy()
                layer = self.add_dense_layer(
                    _input,
                    layer_weight,
                    layer_bias,
                    self.act_dict["identity"],
                    output,
                    name=f"linear_{i}",
                    **kwargs,
                )
                _input = layer.output
        if self._output is None:
            self._output = layer.output

    def get_error(self):
        if self._has_solution:
            t_in = torch.from_numpy(self.input_values).float()
            t_out = self.predictor.forward(t_in)
            return np.abs(t_out.detach().numpy() - self.output.X)
        raise NoSolution()



# Copyright © 2022 Gurobi Optimization, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Module for formulating a
:external+sklearn:py:class:`sklearn.ensemble.GradientBoostingRegressor`
into a :gurobipy:`model`.
"""

from gurobipy import GRB

from ..modeling import AbstractPredictorConstr
from .decision_tree_regressor import add_decision_tree_regressor_constr
from .skgetter import SKgetter


def add_gradient_boosting_regressor_constr(
    gp_model, gradient_boosting_regressor, input_vars, output_vars=None, **kwargs
):
    """Formulate gradient_boosting_regressor into gp_model.

    The formulation predicts the values of output_vars using input_vars
    according to gradient_boosting_regressor. See our :ref:`User's Guide
    <Gradient Boosting Regression>` for details on the mip formulation used.

    Parameters
    ----------
    gp_model : :gurobipy:`model`
        The gurobipy model where the predictor should be inserted.
    gradient_boosting_regressor : :external+sklearn:py:class:`sklearn.ensemble.GradientBoostingRegressor`
        The gradient boosting regressor to insert as predictor.
    input_vars : :gurobipy:`mvar` or :gurobipy:`var` array like
        Decision variables used as input for gradient boosting regressor in model.
    output_vars : :gurobipy:`mvar` or :gurobipy:`var` array like, optional
        Decision variables used as output for gradient boosting regressor in model.

    Returns
    -------
    GradientBoostingRegressorConstr
        Object containing information about what was added to gp_model to formulate
        gradient_boosting_regressor.

    Note
    ----
    |VariablesDimensionsWarn|

    Also see
    :py:func:`gurobi_ml.sklearn.decision_tree_regressor.add_decision_tree_regressor`
    for specific parameters to model decision tree estimators.
    """
    return GradientBoostingRegressorConstr(
        gp_model, gradient_boosting_regressor, input_vars, output_vars, **kwargs
    )


class GradientBoostingRegressorConstr(SKgetter, AbstractPredictorConstr):
    """Class to model trained
    :external+sklearn:py:class:`sklearn.ensemble.GradientBoostingRegressor`
    with gurobipy.

    |ClassShort|
    """

    def __init__(self, gp_model, predictor, input_vars, output_vars, **kwargs):
        self._output_shape = 1
        self.estimators_ = []
        self._default_name = "gbtree_reg"
        SKgetter.__init__(self, predictor, input_vars)
        AbstractPredictorConstr.__init__(
            self, gp_model, input_vars, output_vars, **kwargs
        )

    def _mip_model(self, **kwargs):
        """Predict output variables y from input variables X using the
        decision tree.

        Both X and y should be array or list of variables of conforming dimensions.
        """
        model = self._gp_model
        predictor = self.predictor

        _input = self._input
        output = self._output
        nex = _input.shape[0]

        outdim = output.shape[1]
        assert (
            outdim == 1
        ), "Output dimension of gradient boosting regressor should be 1"

        estimators = []
        if self._no_debug:
            kwargs["no_record"] = True

        tree_vars = model.addMVar(
            (nex, predictor.n_estimators_, 1),
            lb=-GRB.INFINITY,
            name=self._name_var("esimator"),
        )

        for i in range(predictor.n_estimators_):
            tree = predictor.estimators_[i]
            if self.verbose:
                self._timer.timing(f"Estimator {i}")
            estimators.append(
                add_decision_tree_regressor_constr(
                    model, tree[0], _input, tree_vars[:, i, :], **kwargs
                )
            )
        self.estimators_ = estimators

        constant = predictor.init_.constant_
        model.addConstr(
            output == predictor.learning_rate * tree_vars.sum(axis=1) + constant[0][0]
        )

    def print_stats(self, abbrev=False, file=None):
        """Print statistics on model additions stored by this class.

        This function prints detailed statistics on the variables
        and constraints that where added to the model.

        Includes a summary of the estimators that it contains.

        Arguments
        ---------

        file: None, optional
            Text stream to which output should be redirected. By default sys.stdout.
        """
        super().print_stats(abbrev=abbrev, file=file)
        if abbrev or self._no_debug:
            return
        print(file=file)

        self._print_container_steps("Estimator", self.estimators_, file=file)



# Copyright © 2022 Gurobi Optimization, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Implementation for using sklearn preprocessing object in a
Guobi model.
"""

import gurobipy as gp

from ..exceptions import NoModel
from .skgetter import SKtransformer


def add_polynomial_features_constr(gp_model, polynomial_features, input_vars, **kwargs):
    """Formulate polynomial_features into gp_model.

    Note that this function creates the output variables from
    the input variables.

    Parameters
    ----------
    gp_model : :gurobipy:`model`
        The gurobipy model where polynomial features should be inserted.
    polynomial_features : :external+sklearn:py:class:`sklearn.preprocessing.PolynomialFeatures`
        The polynomial features to insert in gp_model.
    input_vars : :gurobipy:`mvar` or :gurobipy:`var` array like
        Decision variables used as input for polynomial features in model.

    Returns
    -------
    sklearn.preprocessing.PolynomialFeaturesConstr
        Object containing information about what was added to gp_model to insert the
        polynomial_features in it

    Warning
    -------
    Only polynomial features of degree 2 are supported.
    """
    return PolynomialFeaturesConstr(gp_model, polynomial_features, input_vars, **kwargs)


def add_standard_scaler_constr(gp_model, standard_scaler, input_vars, **kwargs):
    """Formulate standard_scaler into gp_model.

    Note that this function creates the output variables from
    the input variables.

    Parameters
    ----------
    gp_model : :gurobipy:`model`
        The gurobipy model where the standard scaler should be inserted.
    standard_scaler : :external+sklearn:py:class:`sklearn.preprocessing.StandardScaler`
        The standard scaler to insert as predictor.
    input_vars : :gurobipy:`mvar` or :gurobipy:`var` array like
        Decision variables used as input for standard scaler in model.

    Returns
    -------
    sklearn.preprocessing.StandardScalerConstr
        Object containing information about what was added to gp_model to insert the
        standard_scaler in it
    """
    return StandardScalerConstr(gp_model, standard_scaler, input_vars, **kwargs)


class StandardScalerConstr(SKtransformer):
    """Class to model a fitted
    :external+sklearn:py:class:`sklearn.preprocessing.StandardScaler` with
    gurobipy.

    Stores the changes to :gurobipy:`model` when embedding an instance into it.
    """

    def __init__(self, gp_model, scaler, input_vars, **kwargs):
        self._default_name = "std_scaler"
        self._output_shape = scaler.n_features_in_
        super().__init__(gp_model, scaler, input_vars, **kwargs)

    def _mip_model(self, **kwargs):
        """Do the transformation on x."""
        _input = self._input
        output = self._output

        scale = self.transformer.scale_
        mean = self.transformer.mean_

        self._gp_model.addConstr(
            _input - output * scale == mean, name=self._name_var("s")
        )
        return self


class PolynomialFeaturesConstr(SKtransformer):
    """Class to model trained
    :external+sklearn:py:class:`sklearn.preprocessing.PolynomialFeatures` with
    gurobipy.
    """

    def __init__(self, gp_model, polynomial_features, input_vars, **kwargs):
        if polynomial_features.degree > 2:
            raise NoModel(
                polynomial_features, "Can only handle polynomials of degree < 2"
            )
        self._default_name = "poly_feat"
        super().__init__(gp_model, polynomial_features, input_vars, **kwargs)

    def _mip_model(self, **kwargs):
        """Do the transformation on x."""
        _input = self._input
        output = self._output

        n_examples, n_feat = _input.shape
        powers = self.transformer.powers_
        assert powers.shape[0] == self.transformer.n_output_features_
        assert powers.shape[1] == n_feat

        for k in range(n_examples):
            for i, power in enumerate(powers):
                q_expr = gp.QuadExpr()
                q_expr += 1.0
                for j, feat in enumerate(_input[k, :]):
                    if power[j] == 2:
                        q_expr *= feat.item()
                        q_expr *= feat.item()
                    elif power[j] == 1:
                        q_expr *= feat.item()
                self.gp_model.addConstr(
                    output[k, i] == q_expr, name=self._indexed_name((k, i), "polyfeat")
                )


def sklearn_transformers():
    """Return dictionary of Scikit Learn preprocessing objects."""
    return {
        "StandardScaler": add_standard_scaler_constr,
        "PolynomialFeatures": add_polynomial_features_constr,
    }



# Copyright © 2022 Gurobi Optimization, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""To get version and git hash of built package."""
from importlib import metadata

try:
    __version__ = metadata.version("gurobi_machinelearning")
except metadata.PackageNotFoundError:
    __version__ = "dev"

GIT_HASH = "411376135a2001b5fcd3f57261b51ab673e94761"


def get_versions():
    """Get package version."""
    # Downloaded package with inserted git hash.
    if "Format" not in GIT_HASH:
        git_hash = f"-{GIT_HASH}"
    # No inserted git hash, the repo is probably cloned.
    else:
        git_hash = ""

    return {"short": __version__, "long": f"{__version__}{git_hash}"}



# Copyright © 2022 Gurobi Optimization, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Define generic function that can add any known trained predictor."""
import sys

from .register_user_predictor import user_predictors


def sklearn_convertors():
    """Collect convertors for Scikit-learn objects."""
    if "sklearn" in sys.modules:
        from .sklearn import (  # pylint: disable=import-outside-toplevel
            add_pipeline_constr,
        )
        from .sklearn.predictors_list import (  # pylint: disable=import-outside-toplevel
            sklearn_predictors,
        )
        from .sklearn.preprocessing import sklearn_transformers

        return (
            sklearn_predictors()
            | sklearn_transformers()
            | {
                "Pipeline": add_pipeline_constr,
            }
        )
    return {}


def pytorch_convertors():
    """Collect known PyTorch objects that can be formulated and the conversion class."""
    if "torch" in sys.modules:
        import torch  # pylint: disable=import-outside-toplevel

        from .torch import (  # pylint: disable=import-outside-toplevel
            add_sequential_constr,
        )

        return {torch.nn.Sequential: add_sequential_constr}
    return {}


def keras_convertors():
    """Collect known Keras objects that can be embedded and the conversion class."""
    if "tensorflow" in sys.modules:
        from keras.engine.functional import (  # pylint: disable=import-outside-toplevel
            Functional,
        )
        from keras.engine.training import (  # pylint: disable=import-outside-toplevel
            Model,
        )
        from tensorflow import keras  # pylint: disable=import-outside-toplevel

        from .keras import add_keras_constr  # pylint: disable=import-outside-toplevel

        return {
            keras.Sequential: add_keras_constr,
            Functional: add_keras_constr,
            Model: add_keras_constr,
        }
    return {}


def registered_predictors():
    """Return the list of registered predictors."""
    convertors = {}
    convertors |= sklearn_convertors()
    convertors |= pytorch_convertors()
    convertors |= keras_convertors()
    convertors |= user_predictors()
    return convertors



# Copyright © 2022 Gurobi Optimization, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Module for formulating a Keras model into a :gurobipy:`model`."""

import numpy as np
from tensorflow import keras

from ..exceptions import NoModel, NoSolution
from ..modeling.neuralnet import BaseNNConstr


def add_keras_constr(gp_model, keras_model, input_vars, output_vars=None, **kwargs):
    """Formulate keras_model into gp_model.

    The formulation predicts the values of output_vars using input_vars according to keras_model.
    See our :ref:`Users Guide <Neural Networks>` for details on the mip formulation used.

    Parameters
    ----------
    gp_model : :gurobipy:`model`
        The gurobipy model where the predictor should be inserted.
    keras_model : `keras.Model <https://keras.io/api/models/model/>`
        The keras model to insert as predictor.
    input_vars : :gurobipy:`mvar` or :gurobipy:`var` array like
        Decision variables used as input for Keras model in gp_model.
    output_vars : :gurobipy:`mvar` or :gurobipy:`var` array like, optional
        Decision variables used as output for Keras model in gp_model.

    Returns
    -------
    KerasNetworkConstr
        Object containing information about what was added to gp_model to formulate
        keras_model into it

    Raises
    ------
    NoModel
        If the translation for some of the Keras model structure
        (layer or activation) is not implemented.

    Warning
    -------

      Only `Dense <https://keras.io/api/layers/core_layers/dense/>`_ (possibly
      with `relu` activation), and `ReLU <https://keras.io/api/layers/activation_layers/relu/>`_ with
      default settings are supported.

    Note
    ----
    |VariablesDimensionsWarn|
    """
    return KerasNetworkConstr(gp_model, keras_model, input_vars, output_vars, **kwargs)


class KerasNetworkConstr(BaseNNConstr):
    """Class to model trained `keras.Model <https://keras.io/api/models/model/>` with gurobipy.

    |ClassShort|
    """

    def __init__(self, gp_model, predictor, input_vars, output_vars=None, **kwargs):
        assert predictor.built
        for step in predictor.layers:
            if isinstance(step, keras.layers.Dense):
                config = step.get_config()
                activation = config["activation"]
                if activation not in ("relu", "linear"):
                    raise NoModel(predictor, f"Unsupported activation {activation}")
            elif isinstance(step, keras.layers.ReLU):
                if step.negative_slope != 0.0:
                    raise NoModel(
                        predictor, "Only handle ReLU layers with negative slope 0.0"
                    )
                if step.threshold != 0.0:
                    raise NoModel(
                        predictor, "Only handle ReLU layers with threshold of 0.0"
                    )
                if step.max_value is not None and step.max_value < float("inf"):
                    raise NoModel(predictor, "Only handle ReLU layers without maxvalue")
            elif isinstance(step, keras.layers.InputLayer):
                pass
            else:
                raise NoModel(
                    predictor, f"Unsupported network layer {type(step).__name__}"
                )

        super().__init__(gp_model, predictor, input_vars, output_vars, **kwargs)

    def _mip_model(self, **kwargs):
        network = self.predictor
        _input = self._input
        output = None
        num_layers = len(network.layers)

        for i, step in enumerate(network.layers):
            if i == num_layers - 1:
                output = self._output
            if isinstance(step, keras.layers.InputLayer):
                pass
            elif isinstance(step, keras.layers.ReLU):
                layer = self.add_activation_layer(
                    _input, self.act_dict["relu"], output, name=f"relu{i}", **kwargs
                )
                _input = layer.output
            else:
                config = step.get_config()
                activation = config["activation"]
                if activation == "linear":
                    activation = "identity"
                weights, bias = step.get_weights()
                layer = self.add_dense_layer(
                    _input,
                    weights,
                    bias,
                    self.act_dict[activation],
                    output,
                    name=f"dense{i}",
                    **kwargs,
                )
                _input = layer.output
        if self._output is None:
            self._output = layer.output

    def get_error(self):
        if self._has_solution:
            return np.abs(self.predictor.predict(self.input_values) - self.output.X)
        raise NoSolution()



# Copyright © 2022 Gurobi Optimization, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Exceptions for gurobi_ml."""


class NotRegistered(Exception):
    """Predictor is not supported by gurobi_ml."""

    def __init__(self, predictor):
        super().__init__(
            f"Object of type {predictor} is not registered/supported with gurobi_ml"
        )


class NoModel(Exception):
    """No model is known for some structure."""

    def __init__(self, predictor, reason):
        if not isinstance(predictor, str):
            predictor = type(predictor).__name__
        super().__init__(f"Can't do model for {predictor}: {reason}")


class NoSolution(Exception):
    """Gurobi doesn't have a solution."""

    def __init__(self):
        super().__init__("No solution available")


class ParameterError(Exception):
    """Wrong parameter to a function."""

    def __init__(self, message):
        super().__init__(message)



# Copyright © 2022 Gurobi Optimization, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Bases classes for modeling neural network layers."""


from .._var_utils import _default_name
from ..base_predictor_constr import AbstractPredictorConstr
from .activations import Identity, ReLU
from .layers import ActivationLayer, DenseLayer


class BaseNNConstr(AbstractPredictorConstr):
    """Base class for inserting a regressor based on neural-network/tensor into Gurobi."""

    def __init__(self, gp_model, predictor, input_vars, output_vars, **kwargs):
        self.predictor = predictor
        self.act_dict = {
            "relu": ReLU(),
            "identity": Identity(),
        }
        try:
            for activation, activation_model in kwargs["activation_models"].items():
                self.act_dict[activation] = activation_model
        except KeyError:
            pass
        self._layers = []

        self._default_name = _default_name(predictor)
        super().__init__(
            gp_model=gp_model,
            input_vars=input_vars,
            output_vars=output_vars,
            **kwargs,
        )

    def __iter__(self):
        return self._layers.__iter__()

    def add_dense_layer(
        self,
        input_vars,
        layer_coefs,
        layer_intercept,
        activation,
        activation_vars=None,
        **kwargs,
    ):
        """Add a dense layer to gurobipy model.

        Parameters
        ----------

        input_vars : mvar_array_like
            Decision variables used as input for predictor in model.
        layer_coefs:
            Coefficient for each node in a layer
        layer_intercept:
            Intercept bias
        activation:
            Activation function
        activation_vars : None, optional
            Output variables
        """
        layer = DenseLayer(
            self._gp_model,
            activation_vars,
            input_vars,
            layer_coefs,
            layer_intercept,
            activation,
            **kwargs,
        )
        self._layers.append(layer)
        return layer

    def add_activation_layer(
        self, input_vars, activation, activation_vars=None, **kwargs
    ):
        """Add an activation layer to gurobipy model.

        Parameters
        ----------

        input_vars : mvar_array_like
            Decision variables used as input for predictor in gurobipy model.
        activation:
            Activation function
        activation_vars : mvar_array_like, optional
            Output variables
        """
        layer = ActivationLayer(
            self._gp_model, activation_vars, input_vars, activation, **kwargs
        )
        self._layers.append(layer)
        return layer

    def print_stats(self, abbrev=False, file=None):
        """Print statistics about submodel created.

        Parameters
        ----------

        file : None, optional
            Text stream to which output should be redirected. By default sys.stdout.
        """
        if not abbrev:
            super().print_stats(abbrev, file)
            print(file=file)

            self._print_container_steps("Layer", self._layers, file=file)
        else:
            for layer in self:
                layer.print_stats(abbrev=True, file=file)
                print(file=file)



# Copyright © 2022 Gurobi Optimization, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Module for formulating a column transformer in a :gurobipy:`gurobipy model <Model>`."""

import gurobipy as gp

from .preprocessing import sklearn_transformers
from .skgetter import SKtransformer


class ColumnTransformerConstr(SKtransformer):
    """Class to model a fitted :external+sklearn:py:class:`sklearn.compose.ColumnTransformer` with gurobipy.

    Note
    ----
    This object differs from all the other object in the Gurobi Machine Learning package in
    that it may not be possible to write all of its input with Gurobi variables. Specifically
    some input may consist of categorical features to be encoded using the column transformer.

    Then such input cannot be directly represented in Gurobi but the result of the encoding may be represented.
    If the categorical features are fixed for the optimization model, we can use it so it is allowed by
    the ColumnTransormerConstr class.

    The rule we use to apply the ColumnTransformer to the input is that if the set of columns to which a preprocessing
    transformation is constant in the input we use directly the scikit learn preprocessing object. It at least one of the columns
    is made of gurobipy variables, we use the gurobi-ml object (if it exists).
    """

    def __init__(self, gp_model, column_transformer, input_vars, **kwargs):
        self._default_name = "col_trans"
        super().__init__(gp_model, column_transformer, input_vars, **kwargs)

    # For this class we need to reimplement submodel because we don't want
    # to transform input variables to Gurobi variable. We can't do it for categorical
    # The input should be unchanged.
    def _build_submodel(self, gp_model, *args, **kwargs):
        """Predict output from input using predictor or transformer."""
        _input = self.input
        if hasattr(self._input, "columns"):
            self._input_columns = _input.columns
        if hasattr(self._input, "index"):
            self._input_index = _input.index
        self._mip_model(**kwargs)
        assert self._output is not None
        return self

    def _mip_model(self, **kwargs):
        """Do the transformation on x."""
        column_transform = self.transformer
        _input = self._input
        transformers = {k.lower(): v for k, v in sklearn_transformers().items()}
        transformed = []
        for name, trans, cols in column_transform.transformers_:
            if trans == "passthrough":
                transformed.append(_input.loc[:, cols])
            elif trans == "drop":
                pass
            else:
                data = _input.loc[:, cols]
                anyvar = any(
                    map(lambda i: isinstance(i, gp.Var), data.to_numpy().ravel())
                )
                if anyvar:
                    if name in transformers:
                        trans_constr = transformers[name](self._gp_model, trans, data)
                    transformed.append(trans_constr.output.tolist())
                else:
                    transformed.append(trans.transform(_input.loc[:, cols]))
        self._output = column_transform._hstack(transformed)


def add_column_transformer_constr(gp_model, column_transformer, input_vars, **kwargs):
    """Formulate column_transformer in gurobipy model.

    Parameters
    ----------
    gp_model : :gurobipy:`model`
        The gurobipy model where polynomial features should be inserted.
    column_transformer : :external+sklearn:py:class:`sklearn.compose.ColumnTransformer`
        The column transformer to insert in gp_model.
    input_vars : :gurobipy:`mvar` or :gurobipy:`var` array like
        Decision variables used as input for polynomial features in model.

    Returns
    -------
    sklearn.preprocessing.ColumnTransformerConstr
        Object containing information about what was added to gp_model to insert the
        polynomial_features in it.
    """
    return ColumnTransformerConstr(gp_model, column_transformer, input_vars, **kwargs)



# Copyright © 2022 Gurobi Optimization, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Module for formulating a
:external+sklearn:py:class:`sklearn.tree.DecisionTreeRegressor`
in a :gurobipy:`model`.
"""


import numpy as np
from gurobipy import GRB

from ..modeling import AbstractPredictorConstr
from .skgetter import SKgetter


def add_decision_tree_regressor_constr(
    gp_model,
    decision_tree_regressor,
    input_vars,
    output_vars=None,
    epsilon=0.0,
    **kwargs,
):
    """Formulate decision_tree_regressor into gp_model.

    The formulation predicts the values of output_vars using input_vars
    according to decision_tree_regressor. See our :ref:`User's Guide <Decision
    Tree Regression>` for details on the mip formulation used.

    Parameters
    ----------
    gp_model : :gurobipy:`model`
        The gurobipy model where the predictor should be inserted.
    decision_tree_regressor : :external+sklearn:py:class:`sklearn.tree.DecisionTreeRegressor`
        The decision tree regressor to insert as predictor.
    input_vars : :gurobipy:`mvar` or :gurobipy:`var` array like
        Decision variables used as input for decision tree in model.
    output_vars : :gurobipy:`mvar` or :gurobipy:`var` array like, optional
        Decision variables used as output for decision tree in model.
    epsilon : float, optional
        Small value used to impose strict inequalities for splitting nodes in
        MIP formulations.
    Returns
    -------
    DecisionTreeRegressorConstr
        Object containing information about what was added to gp_model to
        formulate decision_tree_regressor

    Note
    ----

    |VariablesDimensionsWarn|

    Warning
    -------

    Although decision trees with multiple outputs are tested they were never
    used in a non-trivial optimization model. It should be used with care at
    this point.
    """
    return DecisionTreeRegressorConstr(
        gp_model, decision_tree_regressor, input_vars, output_vars, epsilon, **kwargs
    )


class DecisionTreeRegressorConstr(SKgetter, AbstractPredictorConstr):
    """Class to model trained
    :external+sklearn:py:class:`sklearn.tree.DecisionTreeRegressor` with
    gurobipy.

    |ClassShort|
    """

    def __init__(
        self,
        gp_model,
        predictor,
        input_vars,
        output_vars=None,
        epsilon=0.0,
        scale=1.0,
        float_type=np.float32,
        formulation="leafs",
        **kwargs,
    ):
        self.epsilon = epsilon
        self.scale = scale
        self.float_type = float_type
        self._default_name = "tree_reg"

        formulations = ("leafs", "paths")
        if formulation not in formulations:
            raise ValueError(
                "Wrong value for formulation should be one of {}.".format(formulations)
            )
        self._formulation = formulation
        SKgetter.__init__(self, predictor, input_vars)
        AbstractPredictorConstr.__init__(
            self, gp_model, input_vars, output_vars, **kwargs
        )

    def _compute_leafs_bounds(self):

        tree = self.predictor.tree_

        node_lb = -np.ones((tree.n_features, tree.capacity)) * GRB.INFINITY
        node_ub = np.ones((tree.n_features, tree.capacity)) * GRB.INFINITY

        children_left = tree.children_left
        children_right = tree.children_right
        feature = tree.feature
        threshold = tree.threshold

        stack = [
            0,
        ]
        while len(stack):
            node = stack.pop()
            left = children_left[node]
            if left < 0:
                continue
            right = children_right[node]
            assert left not in stack
            assert right not in stack
            node_ub[:, right] = node_ub[:, node]
            node_lb[:, right] = node_lb[:, node]
            node_ub[:, left] = node_ub[:, node]
            node_lb[:, left] = node_lb[:, node]

            node_ub[feature[node], left] = threshold[node]
            node_lb[feature[node], right] = threshold[node] + self.epsilon
            stack.append(right)
            stack.append(left)
        return (node_lb, node_ub)

    def _leaf_mip_model(self, **kwargs):
        tree = self.predictor.tree_
        model = self._gp_model

        _input = self._input
        output = self._output
        outdim = output.shape[1]
        nex = _input.shape[0]

        verbose = self.verbose

        timer = AbstractPredictorConstr._ModelingTimer()

        # Collect leaf nodes
        leafs = tree.children_left < 0
        leafs_vars = model.addMVar(
            (nex, sum(leafs)), vtype=GRB.BINARY, name=self._name_var("leafs")
        )

        if verbose:
            timer.timing(f"Added {nex*sum(leafs)} leafs vars")
        (node_lb, node_ub) = self._compute_leafs_bounds()
        input_ub = _input.getAttr(GRB.Attr.UB)
        input_lb = _input.getAttr(GRB.Attr.LB)

        for i, node in enumerate(leafs.nonzero()[0]):
            reachable = (input_ub >= node_lb[:, node]).all(axis=1) & (
                input_lb <= node_ub[:, node]
            ).all(axis=1)
            # Non reachable nodes
            leafs_vars[~reachable, i].setAttr(GRB.Attr.UB, 0.0)
            # Leaf node:
            rhs = output[reachable, :].tolist()
            lhs = leafs_vars[reachable, i].tolist()
            values = tree.value[node, :, 0]
            n_indicators = sum(reachable)
            for l_var, r_vars in zip(lhs, rhs):
                for r_var, value in zip(r_vars, values):
                    model.addGenConstrIndicator(l_var, 1, r_var, GRB.EQUAL, value)

            for feature in range(tree.n_features):
                lb = node_lb[feature, node]
                ub = node_ub[feature, node]

                if lb > -GRB.INFINITY:
                    tight = (input_lb[:, feature] < lb) & reachable
                    lhs = leafs_vars[tight, i].tolist()
                    rhs = _input[tight, feature].tolist()
                    n_indicators += sum(tight)
                    for l_var, r_var in zip(lhs, rhs):
                        model.addGenConstrIndicator(
                            l_var, 1, r_var, GRB.GREATER_EQUAL, lb
                        )

                if ub < GRB.INFINITY:
                    tight = (input_ub[:, feature] > ub) & reachable
                    lhs = leafs_vars[tight, i].tolist()
                    rhs = _input[tight, feature].tolist()
                    n_indicators += sum(tight)
                    for l_var, r_var in zip(lhs, rhs):
                        model.addGenConstrIndicator(l_var, 1, r_var, GRB.LESS_EQUAL, ub)
            if verbose:
                timer.timing(f"Added leaf {node} using {n_indicators} indicators")

        # We should attain 1 leaf
        model.addConstr(leafs_vars.sum(axis=1) == 1)

        if verbose:
            timer.timing(f"Added {nex} linear constraints")

        output.setAttr(GRB.Attr.LB, np.min(tree.value))
        output.setAttr(GRB.Attr.UB, np.max(tree.value))

    def _paths_mip_model(self, **kwargs):
        tree = self.predictor.tree_
        model = self._gp_model

        _input = self._input
        output = self._output
        outdim = output.shape[1]
        nex = _input.shape[0]
        nodes = model.addMVar(
            (nex, tree.capacity), vtype=GRB.BINARY, name=self._name_var("node")
        )

        # Collect leafs and non-leafs nodes
        notleafs = tree.children_left >= 0
        leafs = tree.children_left < 0

        # Connectivity constraint
        model.addConstr(
            nodes[:, notleafs]
            == nodes[:, tree.children_right[notleafs]]
            + nodes[:, tree.children_left[notleafs]]
        )

        # The value of the root is always 1
        nodes[:, 0].LB = 1.0

        # Node splitting
        for node in notleafs.nonzero()[0]:
            left = tree.children_left[node]
            right = tree.children_right[node]
            threshold = tree.threshold[node]
            threshold = self.float_type(threshold)
            scale = max(abs(1 / threshold), self.scale)
            # Intermediate node
            feature = tree.feature[node]
            feat_var = _input[:, feature]

            fixed_input = (feat_var.UB == feat_var.LB).all()

            if fixed_input:
                # Special case where we have an MVarPlusConst object
                # If that feature is a constant we can directly fix it.
                value = _input[:, feature].LB
                fixed_left = value <= threshold
                nodes[fixed_left, right].UB = 0.0
                nodes[~fixed_left, left].UB = 0.0
            else:
                lhs = _input[:, feature].tolist()
                rhs = nodes[:, left].tolist()
                threshold *= scale
                model.addConstrs(
                    ((rhs[k] == 1) >> (scale * lhs[k] <= threshold)) for k in range(nex)
                )
                rhs = nodes[:, right].tolist()
                model.addConstrs(
                    ((rhs[k] == 1) >> (scale * lhs[k] >= threshold + self.epsilon))
                    for k in range(nex)
                )

        for node in leafs.nonzero()[0]:
            # Leaf node:
            lhs = output.tolist()
            rhs = nodes[:, node].tolist()
            value = tree.value[node, :, 0]
            model.addConstrs(
                (rhs[k] == 1) >> (lhs[k][i] == value[i])
                for k in range(nex)
                for i in range(outdim)
            )

        output.LB = np.min(tree.value)
        output.UB = np.max(tree.value)

    def _mip_model(self, **kwargs):
        if self._formulation == "leafs":
            return self._leaf_mip_model(**kwargs)
        else:
            return self._paths_mip_model(**kwargs)



# Copyright © 2022 Gurobi Optimization, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Building Sub-models with gurobipy."""

from time import time


class SubModel:
    """Base class for building and representing a MIP formulation embedded in a gurobipy.Model.

    When instantiating this class, a (sub-)model is created in the provided
    gurobipy.Model.  The instance represents the sub-model that was created,
    and can be removed from it again.  Furthermore, additional information
    about the submodel can be provided, depending on the sub-model creation.

    There are two ways of using this class:

    1. To build a derived class to build a specific model.

    For this, you must create a sub-class that defines the _build_submodel()
    method, where you implement the code that builds the sub-model from
    the provided data in the gurobipy.Model.  The class should be implemented
    in the following way:

    class MySubModel(SubModel):
        def __init__(self, model, ...., **kwargs):
            super().__init(self, model, ...., **kwargs)

        def _build_submodel(self, model, ...., **kwargs):
            ...

    Here "...." can be your desired list of parameters, including possibly
    *args.

    To create your sub-model in a gurobipy.Model object, simply instantiate
    your class:

    my_sub_model = MySubModel(model, ....)

    2. To build a sub-model using a modeling function.

    If you have a function that populates a gurobipy.Model with a sub-model
    from data, you can use this class directly to represent the created
    submodel in the SubModel instance.  The signature of such a modeling
    function must be:

    def my_modeling_function(model, ...., **kwargs):
        ...

    Again, "...." can be your desired list of parameters, including possibly
    *args.

    To use your model building function, e.g. my_modeling_function, with this
    class, simply pass it to the SubModel constructor:

    my_sub_model = SubModel(model, ....., model_function=my_modeling_function, ....)

    Independently of how you created your SubModel instance, you can remove
    the sub-model from the gurobipy.Model by calling the remove method:

    my_sub_model.remove()

    You can always pass the named parameter name='my_name' to the constructors
    of a submodel.  This name will be used to prefix the names of all the
    modeling objects that are being created during submodel construction. If
    no name is provided an automatic name is generated from a default name.
    The default may be already specified as an attribute (mostly
    useful with sub-classing); if not, it is derived from the class name of
    your derived class, or the name of your model construction function.
    The only exception is the empty string name='' which is not used as a
    prefix.  This is useful if the full model is built as a SubModel.

    Note, that you can instantiate as many SubModels as needed in one
    gurobipy.Model.  In particular, you can instantiate SubModels from
    within the model construction method or function of a SubModel itself.
    In this case, the automatic name handling will '.'-separated generate
    names following your call hierarchy.

    Parameters
    ----------
    """

    def __init__(
        self,
        gp_model,
        *args,
        model_function=None,
        name=None,
        **kwargs,
    ):
        self._gp_model = None
        self._objects = {}
        self._firstvar = None
        self._lastvar = None
        self._firstconstr = None
        self._lastconstr = None
        self._qconstrs = []
        self._genconstrs = []
        self._sos = []
        self._first_callback = None
        self._last_callback = None
        self._name = None
        self._model_function = model_function

        if not hasattr(self, "_default_name"):
            self._default_name = type(self).__name__

        self.verbose = False
        if "verbose" in kwargs:
            self.verbose = kwargs["verbose"]
            self._timer = SubModel._ModelingTimer()
            print()
            self._timer.timing(f"Start building formulation for {self._default_name}")

        self._no_recording = False
        if "no_record" in kwargs:
            self._no_recording = kwargs["no_record"]

        self._no_debug = False
        if "no_debug" in kwargs:
            self._no_debug = kwargs["no_debug"]
        before = self._open(gp_model)

        if self.verbose:
            self._timer.timing("Model statistics")

        self._objects = self._build_submodel(gp_model, *args, **kwargs)

        if self.verbose:
            self._timer.timing("Built model")

        if self._objects is None:
            self._objects = {}

        self._close(before, name)

        if self.verbose:
            print()

    def _build_submodel(self, gp_model, *args, **kwargs):
        """Method to be overridden for generating the model in a sub-class.

        When using SubModel to wrap a modeling function in a SubModel, the default
        implementation of this method simply calls the modeling function.
        """
        return self._model_function(gp_model, *args, **kwargs)

    class _ModelingTimer:
        def __init__(self):
            self.start = time()
            self.last = self.start

        def timing(self, message: str):
            current = time()
            print(
                f"{message} in {current - self.last:.2f} secs (total: {current - self.start:.2f})."
            )
            self.last = current

    class _ModelingData:
        """Class for recording modeling data in a gurobipy.Model object."""

        def __init__(self):
            self.name_handler = None

        def pop_name_handler(self):
            """Remove name handler and return it."""
            name_handler = self.name_handler
            self.name_handler = None
            return name_handler

        def push_name_handler(self, name_handler):
            """Install name handler."""
            self.name_handler = name_handler

    class _modelstats:
        """Helper class for recording gurobi model dimensions.

        Parameters
        ----------
        gp_model : gp.Model <https://www.gurobi.com/documentation/9.5/refman/py_model.html>
            A gurobipy model

        Attributes
        ----------
        numvars : int
            Number of variables in `gp_model`.
        numconstrs : int
            Number of constraints in `gp_model`.
        numsos : int
            Number of SOS constraints in `gp_model`
        numqconstrs : int
            Number of quadratic constraints in `gp_model`
        numgenconstrs : int
            Number of general constraints in `gp_model`
        """

        def __init__(self, gp_model):
            gp_model.update()
            self.numvars = gp_model.numVars
            self.numconstrs = gp_model.numConstrs
            self.numsos = gp_model.numSOS
            self.numqconstrs = gp_model.numQConstrs
            self.numgenconstrs = gp_model.numGenConstrs

    def _record(self, gp_model, before):
        """Record added modeling objects compared to status before."""
        gp_model.update()
        # range of variables
        if gp_model.numvars > before.numvars:
            self._firstvar = gp_model.getVars()[before.numvars]
            self._lastvar = gp_model.getVars()[gp_model.numvars - 1]
        else:
            self._firstvar = None
            self._lastvar = None
        # range of constraints
        if gp_model.numconstrs > before.numconstrs:
            self._firstconstr = gp_model.getConstrs()[before.numconstrs]
            self._lastconstr = gp_model.getConstrs()[gp_model.numconstrs - 1]
        else:
            self._firstconstr = None
            self._lastconstr = None
        # range of Q constraints
        if gp_model.numqconstrs > before.numqconstrs:
            self._qconstrs = gp_model.getQConstrs()[
                before.numqconstrs : gp_model.numqconstrs
            ]
        else:
            self._qconstrs = []
        # range of GenConstrs
        if gp_model.numgenconstrs > before.numgenconstrs:
            self._genconstrs = gp_model.getGenConstrs()[
                before.numgenconstrs : gp_model.numgenconstrs
            ]
        else:
            self._genconstrs = []
        # range of SOS
        if gp_model.numsos > before.numsos:
            self._sos = gp_model.getSOSs()[before.numsos : gp_model.numsos]
        else:
            self._sos = []

    @property
    def vars(self):
        """Return the list of variables in the submodel."""
        if self._firstvar:
            return self._gp_model.getVars()[
                self._firstvar.index : self._lastvar.index + 1
            ]
        return []

    @property
    def constrs(self):
        """Return the list of linear constraints in the submodel."""
        if self._firstconstr:
            return self._gp_model.getConstrs()[
                self._firstconstr.index : self._lastconstr.index + 1
            ]
        return []

    @property
    def qconstrs(self):
        """Return the list of quadratic constraints in the submodel."""
        return self._qconstrs

    @property
    def genconstrs(self):
        """Return the list of general constraints in the submodel."""
        return self._genconstrs

    @property
    def sos(self):
        """Return the list of SOS constraints in the submodel."""
        return self._sos

    def _open(self, gp_model):
        """Start registering modeling object that are added to the gurobipy.Model."""
        self._gp_model = gp_model
        try:
            modeling_data = gp_model._modeling_data
        except AttributeError:
            modeling_data = SubModel._ModelingData()
        gp_model._modeling_data = modeling_data

        if self._no_recording:
            modelstats = None
        else:
            modelstats = self._modelstats(gp_model)
        return (modelstats, gp_model._modeling_data.pop_name_handler())

    def _close(self, before, name):
        """Finalize addition of modeling objects to the gurobipy.Model object.

        Record all added modeling objects in SubModel and prefix their names.
        """

        class NameHandler:
            """Handle automatic name generation in gp.Model."""

            def __init__(self):
                self.name = {}

            def get_name(self, sub: SubModel, name: str):
                """Return a default name for specified submodel sub."""
                if name is None or name == "":
                    name = sub.default_name
                try:
                    num = self.name[name]
                    self.name[name] = num + 1
                    return f"{name}{num}"
                except KeyError:
                    self.name[name] = 0
                    return name

        def prefix_names(gp_model, objs, attr, name):
            """Prefix all modeling object names with name."""
            if len(objs) == 0:
                return
            object_names = gp_model.getAttr(attr, objs)
            new_names = [f"{name}.{obj_name}" for obj_name in object_names]
            gp_model.setAttr(attr, objs, new_names)

        # re-install name handler
        self._gp_model._modeling_data.push_name_handler(before[1])

        if self._no_recording:
            return

        # record all newly added modeling objects
        self._record(self._gp_model, before[0])

        name_handler = self._gp_model._modeling_data.name_handler
        if name_handler is None:
            name_handler = NameHandler()
            self._gp_model._modeling_data.push_name_handler(name_handler)

        if name != "":
            # prefix names of newly created modeling objects
            name = name_handler.get_name(self, name)
            self._name = name
            prefix_names(self._gp_model, self.vars, "VarName", name)
            prefix_names(self._gp_model, self.constrs, "ConstrName", name)
            prefix_names(self._gp_model, self.qconstrs, "QCName", name)
            prefix_names(self._gp_model, self.genconstrs, "GenConstrName", name)
            # SOS can't have a name! :-O
            # prefix_names(self._gp_model, self.sos, "SOSName", name)
        else:
            name = name_handler.get_name(self, name)
        if self.verbose:
            self._timer.timing(f"Added {name}")

    def print_stats(self, abbrev=False, file=None):
        """Print statistics about submodel.

        This functions prints detailed statistics on the variables
        and constraints that where added to the gp_model using this object.

        Usually derived class reimplement this function to provide more
        details about the structure of the additions (type of ML model,
        layers if it's a neural network,...)

        Parameters
        ----------

        file : None, optional
            Text stream to which output should be redirected. By default sys.stdout.
        """
        if abbrev:
            return
        name = self._name
        print(f"Model for {name}:", file=file)
        print(f"{len(self.vars)} variables", file=file)
        print(f"{len(self.constrs)} constraints", file=file)
        qconstr = len(self.qconstrs)
        if qconstr > 0:
            print(f"{qconstr} quadratic constraints", file=file)
        genconstr = len(self.genconstrs)
        if genconstr > 0:
            print(f"{genconstr} general constraints", file=file)
        sosconstr = len(self.sos)
        if sosconstr > 0:
            print(f"{sosconstr} SOS constraints", file=file)

    @property
    def gp_model(self):
        """Access gurobipy model the submodel is a part of."""
        return self._gp_model

    @property
    def default_name(self):
        """Access the default name base used for automatic name generation.

        :meta private:
        """
        return self._default_name

    def remove(self):
        """Remove the submodel from the model."""
        if self._gp_model:
            self._gp_model.remove(self.vars)
            self._gp_model.remove(self.constrs)
            self._gp_model.remove(self.qconstrs)
            self._gp_model.remove(self.genconstrs)
            self._gp_model.remove(self.sos)
            if self._first_callback:
                beg = self._gp_model.getCallbackIndex(self._first_callback)
                end = self._gp_model.getCallbackIndex(self._last_callback)
                self._gp_model.removeCallbacks(beg, end)

            self._firstvar = None
            self._lastvar = None
            self._firstconstr = None
            self._lastconstr = None
            self._qconstrs = []
            self._genconstrs = []
            self._sos = []
            self._first_callback = None
            self._last_callback = None
            self._gp_model = None
            self._objects = None



# Copyright © 2022 Gurobi Optimization, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Internal module to make MIP modeling of activation functions."""

import numpy as np
from gurobipy import GRB


class Identity:
    """Class to apply identity activation on a neural network layer.

    Parameters
    ----------
    setbounds : Bool
        Optional flag not to set bounds on the output variables.

    Attributes
    ----------
    setbounds : Bool
        Optional flag not to set bounds on the output variables.
    """

    def __init__(self):
        pass

    def mip_model(self, layer):
        """MIP model for identity activation on a layer.

        Parameters
        ----------
        layer : AbstractNNLayer
            Layer to which activation is applied.
        """
        output = layer.output
        layer.gp_model.addConstr(output == layer.input @ layer.coefs + layer.intercept)


class ReLU:
    """Class to apply the ReLU activation on a neural network layer.

    Parameters
    ----------
    setbounds : Bool
        Optional flag not to set bounds on the output variables.
    bigm : Float
        Optional maximal value for bounds use in the formulation

    Attributes
    ----------
    setbounds : Bool
        Optional flag not to set bounds on the output variables.
    bigm : Float
        Optional maximal value for bounds use in the formulation
    """

    def __init__(self):
        pass

    def mip_model(self, layer):
        """MIP model for ReLU activation on a layer.

        Parameters
        ----------
        layer : AbstractNNLayer
            Layer to which activation is applied.
        """
        output = layer.output
        if hasattr(layer, "coefs"):
            if not hasattr(layer, "mixing"):
                mixing = layer.gp_model.addMVar(
                    output.shape,
                    lb=-GRB.INFINITY,
                    vtype=GRB.CONTINUOUS,
                    name=layer._name_var("mix"),
                )
                layer.mixing = mixing
            layer.gp_model.update()

            layer.gp_model.addConstr(
                layer.mixing == layer.input @ layer.coefs + layer.intercept
            )
        else:
            mixing = layer._input
        for index in np.ndindex(output.shape):
            layer.gp_model.addGenConstrMax(
                output[index],
                [
                    mixing[index],
                ],
                constant=0.0,
                name=layer._indexed_name(index, "relu"),
            )
